package bzr4idea;

import static bzr4idea.Utils.root;
import com.intellij.openapi.progress.ProgressIndicator;
import com.intellij.openapi.vcs.FilePath;
import com.intellij.openapi.vcs.FileStatus;
import com.intellij.openapi.vcs.VcsException;
import com.intellij.openapi.vcs.changes.Change;
import com.intellij.openapi.vcs.changes.ChangeProvider;
import com.intellij.openapi.vcs.changes.ChangelistBuilder;
import com.intellij.openapi.vcs.changes.ContentRevision;
import com.intellij.openapi.vcs.changes.CurrentContentRevision;
import com.intellij.openapi.vcs.changes.VcsDirtyScope;
import com.intellij.openapi.vfs.LocalFileSystem;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.vcsUtil.VcsUtil;
import org.vcs.bazaar.client.BazaarRevision;
import org.vcs.bazaar.client.BazaarStatusKind;
import org.vcs.bazaar.client.IBazaarStatus;
import org.vcs.bazaar.client.IBazaarClient;
import org.vcs.bazaar.client.core.BazaarClientException;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * @author Patrik Beno
 */
public class BzrChangeProvider implements ChangeProvider {

	private BzrVcs bzr;
	private boolean ignoredFilesProcessed = false;

	private Map<String, Change> changesCache = new HashMap<String, Change>();

	public BzrChangeProvider(BzrVcs bzr) {
		this.bzr = bzr;
	}

	public void getChanges(VcsDirtyScope vcsDirtyScope, final ChangelistBuilder changelistBuilder,
								  ProgressIndicator progressIndicator) throws VcsException {

//		try {
//			FileDocumentManager.getInstance().saveAllDocuments();
//		} catch (Throwable t) {
//			// this causes exception in some scenarios (seems not always)
//			//		Write access is allowed from event dispatch thread only
//			//		Details: Current thread: Thread[Change List Updater,6,Idea Thread Group] 14958507
//			//		Our dispatch thread:Thread[AWT-EventQueue-1,6,Idea Thread Group] 11203538
//			//		SystemEventQueueThread: Thread[AWT-EventQueue-1,6,Idea Thread Group] 11203538
//			Log.info("BzrChangeProvider.getChanges(): saveAllDocuments() failed: %s", t);
//		}


		Log.info("affected roots: %s%n", new ArrayList(vcsDirtyScope.getAffectedContentRoots()));

		Set<FilePath> files = getAllFilesInDirtyScope(vcsDirtyScope);
		Map<VirtualFile, Set<FilePath>> roots = sortByRoots(files);

		for (Set<FilePath> paths : roots.values()) {

			checkCache(changelistBuilder, files);

			try {
				FilePath path = paths.iterator().next();
				FilePath root = VcsUtil.getFilePath(root(path));
				if (root == null) {
					Log.info("Ignoring path %s", path);
					continue;
				}
//				assert root != null : String.format("Could not determine root for %s", paths.iterator().next());

				if (!ignoredFilesProcessed) {
					// use any file to get info about ignored files
					processIgnoredFiles(changelistBuilder, root);
					ignoredFilesProcessed = true;
				}

				try {
					for (IBazaarStatus status : getStatuses(paths, root)) {
						processStatus(changelistBuilder, root, status);
					}
				} finally {
					revisionsByRoot.remove(root.getIOFile());
				}

			} catch (Exception e) {
				throw new VcsException(e);
			}


		}

	}

	public boolean isModifiedDocumentTrackingRequired() {
		return true;
	}

	/**
	 * Clears cache with file statuses
	 *
	 * @param files list of commited files
	 */
	public void onSuccessfulCommit(List<File> files) {
		changesCache.clear();
	}

	private Set<FilePath> getAllFilesInDirtyScope(VcsDirtyScope vcsDirtyScope) {
		Set<FilePath> files = new HashSet<FilePath>();
		for (FilePath filePath : vcsDirtyScope.getDirtyFiles()) {
			files.add(filePath);
		}
		for (FilePath filePath : vcsDirtyScope.getRecursivelyDirtyDirectories()) {
			files.add(filePath);
		}
		return files;
	}

	private Map<VirtualFile, Set<FilePath>> sortByRoots(Set<FilePath> files) {
		Map<VirtualFile, Set<FilePath>> roots = new HashMap<VirtualFile, Set<FilePath>>();

		for (FilePath fpath : files) {
			VirtualFile root = VcsUtil.getVcsRootFor(bzr.getProject(), fpath);
			Set<FilePath> paths = roots.get(root);
			if (paths == null) {
				paths = new HashSet<FilePath>();
				roots.put(root, paths);
			}
			paths.add(fpath);
		}
		return roots;
	}

	private void checkCache(ChangelistBuilder changelistBuilder, Set<FilePath> files) {
		// try to find file info in cache
		Iterator<FilePath> iterator = files.iterator();
		while (iterator.hasNext()) {
			FilePath path = iterator.next();

			Change cacheChange = changesCache.get(path.getPath());
			if (cacheChange != null) {
				processChange(changelistBuilder, cacheChange);
				iterator.remove();
			}
		}
	}

	private IBazaarStatus[] getStatuses(Set<FilePath> paths,
													FilePath root) throws BazaarClientException {

		bzr.getBzrClient().setWorkDir(root.getIOFile());
		IBazaarStatus[] allstatuses = bzr.getBzrClient().status(new File[]{ new File(".") });
		return allstatuses;
	}

	private void processStatus(ChangelistBuilder changelistBuilder, FilePath root, IBazaarStatus status) {
		Log.info("processStatus() for %s %s", root, status);
		FilePath fpath = VcsUtil.getFilePath(new File(root.getIOFile(), status.getPath()));
		if (status.contains(BazaarStatusKind.CREATED)) {
			Change change = new Change(null, CurrentContentRevision.create(fpath), FileStatus.ADDED);
			processChange(changelistBuilder, fpath, change);
		}
		if (status.contains(BazaarStatusKind.DELETED)) {
			Change change = new Change(getLastContentRevision(fpath), null, FileStatus.DELETED);
			processChange(changelistBuilder, fpath, change);
		}
		if (status.contains(BazaarStatusKind.MODIFIED)) {
			Change change = new Change(getLastContentRevision(fpath), CurrentContentRevision.create(fpath), FileStatus.MODIFIED);
			processChange(changelistBuilder, fpath, change);
		}
		if (status.contains(BazaarStatusKind.RENAMED)) {
			ContentRevision currentRevision = CurrentContentRevision.create(fpath);
			Change change = new Change(getLastContentRevision(VcsUtil.getFilePath(new File(root.getIOFile(), status.getPreviousPath()))), currentRevision, FileStatus.MODIFIED);
			processChange(changelistBuilder, fpath, change);
		}
		if (status.contains(BazaarStatusKind.UNKNOWN)) {
			ContentRevision currentRevision = CurrentContentRevision.create(fpath);
			Change change = new Change(null, currentRevision, FileStatus.UNKNOWN);
			processChange(changelistBuilder, fpath, change);

		}
		if (status.contains(BazaarStatusKind.VERSIONED)) {
			ContentRevision currentRevision = CurrentContentRevision.create(fpath);
			Change change = new Change(currentRevision, currentRevision, FileStatus.NOT_CHANGED);
			processChange(changelistBuilder, fpath, change);

		}
	}

	private void processChange(ChangelistBuilder changelistBuilder, FilePath fpath, Change change) {
		processChange(changelistBuilder, change);
		changesCache.put(fpath.getPath(), change);
	}

	private void processChange(ChangelistBuilder changelistBuilder, Change change) {
		if (change.getFileStatus() == FileStatus.UNKNOWN) {
			String unknownFilePath = change.getAfterRevision().getFile().getPath();
			VirtualFile virtualFile = LocalFileSystem.getInstance().findFileByPath(unknownFilePath);

			changelistBuilder.processUnversionedFile(virtualFile);
		} else if (change.getFileStatus() == FileStatus.IGNORED) {
			String ignored = change.getAfterRevision().getFile().getPath();
			VirtualFile virtualFile = LocalFileSystem.getInstance().findFileByPath(ignored);

			changelistBuilder.processIgnoredFile(virtualFile);
		} else {
			changelistBuilder.processChange(change);
		}
	}

	private void processIgnoredFiles(ChangelistBuilder changelistBuilder, FilePath anyFile) throws Exception {
		File root = root(anyFile.getIOFile());
		bzr.getBzrClient().setWorkDir(root);
		Map<String, String> map = bzr.getBzrClient().ignored(root);
		for (String ignored : map.keySet()) {
			String ignoredPath = new File(root, ignored).getPath();
			ContentRevision currentRevision = CurrentContentRevision
					.create(VcsUtil.getFilePath(ignoredPath));
			Change change = new Change(null, currentRevision, FileStatus.IGNORED);
			processChange(changelistBuilder, change);
			changesCache.put(ignoredPath, change);
		}
	}

	private ContentRevision getLastContentRevision(FilePath file) {
		try {
			File root = root(file.getIOFile());
			BazaarRevision rev = getRevision(root);
			return new BzrContentRevision(bzr, file.getIOFile(), rev.getValue());
		} catch (BazaarClientException e) {
			throw Assert.notYetHandled(e);
		}
	}

	class RevMap extends HashMap<FilePath,BazaarRevision> {}

	Map<File,BazaarRevision> revisionsByRoot = new HashMap<File, BazaarRevision>();

	BazaarRevision getRevision(File root) throws BazaarClientException {
		BazaarRevision rev = revisionsByRoot.get(root);
		if (rev != null) { return rev; }

		IBazaarClient client = bzr.getBzrClient();
		client.setWorkDir(root);
		rev = client.revno(root);

		revisionsByRoot.put(root, rev);
		return rev;
	}

}
