package bzr4idea;

import com.intellij.ide.startup.StartupActionScriptManager;
import com.intellij.openapi.vcs.AbstractVcsHelper;
import com.intellij.openapi.vcs.VcsShowConfirmationOption;
import com.intellij.openapi.vcs.AbstractVcs;
import com.intellij.openapi.vcs.ProjectLevelVcsManager;
import com.intellij.openapi.vfs.LocalFileOperationsHandler;
import com.intellij.openapi.vfs.LocalFileSystem;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.project.ProjectManager;
import org.jetbrains.annotations.Nullable;
import org.vcs.bazaar.client.commandline.syntax.IRemoveOptions;
import org.vcs.bazaar.client.core.BazaarClientException;
import static org.vcs.bazaar.client.utils.BazaarUtilities.getRootBranch;

import java.io.File;
import java.io.IOException;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;


/**
 * @author Patrik Beno
 */
public class BzrLocalFileOperationsHandler implements LocalFileOperationsHandler {

	public boolean delete(VirtualFile file) throws IOException {
		BzrVcs bzr = getVCS(file);
		if (bzr == null) { return false; }

		try {
			File f = new File(file.getPath());
			bzr.getBzrClient().setWorkDir(getRootBranch(f));
			bzr.getBzrClient().remove(new File[]{f}, IRemoveOptions.FORCE);
			return true;
		} catch (BazaarClientException e) {
			throw Assert.notYetHandled(e);
		}
	}

	public boolean move(VirtualFile srcfile, VirtualFile todir) throws IOException {
		BzrVcs bzr = getVCS(srcfile);
		if (bzr == null) { return false; }

		try {
			File src = new File(srcfile.getPath());
			File dst = new File(new File(todir.getPath()), srcfile.getName());
			bzr.getBzrClient().setWorkDir(getRootBranch(src));
			bzr.getBzrClient().move(src, dst);
			return true;
		} catch (BazaarClientException e) {
			throw Assert.notYetHandled(e);
		}
	}

	@Nullable
	public File copy(VirtualFile srcfile, VirtualFile dstdir, String name) throws IOException {
		BzrVcs bzr = getVCS(srcfile);
		if (bzr == null) { return null; }

		File src = new File(srcfile.getPath());
		File dst = new File(dstdir.getPath(), name);

		// todo: find a better way to copy files
		StartupActionScriptManager.CopyCommand copyCommand =
				new StartupActionScriptManager.CopyCommand(src, dst);
		copyCommand.execute();

		addFileUnderVcs(bzr, dst);

		return dst;
	}

	public boolean rename(VirtualFile file, String newname) throws IOException {
		BzrVcs bzr = getVCS(file);
		if (bzr == null) { return false; }

		try {
			File orig = new File(file.getPath());
			File newFile = new File(file.getParent().getPath(), newname);
			bzr.getBzrClient().setWorkDir(getRootBranch(orig));
			bzr.getBzrClient().move(orig, newFile);
			return true;
		} catch (BazaarClientException e) {
			throw Assert.notYetHandled(e);
		}
	}

	public boolean createFile(VirtualFile fdir, String name) throws IOException {
		BzrVcs bzr = getVCS(fdir);
		if (bzr == null) { return false; }

		File f = new File(fdir.getPath(), name);
		f.createNewFile();
		addFileUnderVcs(bzr, f);
		return true;
	}

	public boolean createDirectory(VirtualFile dir, String name) throws IOException {
		BzrVcs bzr = getVCS(dir);
		if (bzr == null) { return false; }

		File f = new File(dir.getPath(), name);
		f.mkdirs();
		addFileUnderVcs(bzr, f);
		return true;
	}

	private void addFileUnderVcs(BzrVcs bzr, File dstFile) throws IOException {
		VcsShowConfirmationOption.Value confirmationValue = bzr.getAddConfirmation().getValue();

		if (confirmationValue != VcsShowConfirmationOption.Value.DO_NOTHING_SILENTLY) {
			List<VirtualFile> addedVFiles = new LinkedList<VirtualFile>();
			addedVFiles.add(LocalFileSystem.getInstance().refreshAndFindFileByIoFile(dstFile));

			Collection<VirtualFile> filesToProcess;

			if (confirmationValue == VcsShowConfirmationOption.Value.SHOW_CONFIRMATION) {
				final AbstractVcsHelper vcsHelper = AbstractVcsHelper.getInstance(bzr.getProject());

				filesToProcess = vcsHelper.selectFilesToProcess(addedVFiles,
																				"Add files",
																				null,
																				"Add file",
																				"Add file {0} to repository?",
																				bzr.getAddConfirmation());
			} else {
				filesToProcess = addedVFiles;
			}

			if (filesToProcess != null) {
				for (VirtualFile fileToAdd : filesToProcess) {
					try {
						File f = new File(fileToAdd.getPath());
						bzr.getBzrClient().setWorkDir(getRootBranch(f));
						bzr.getBzrClient().add(new File[]{f});

					} catch (Exception e) {
						new File(fileToAdd.getPath()).delete();
						throw new IOException(e);
					}
				}
			}
		}
	}

	@Nullable
	static private BzrVcs getVCS(VirtualFile ... files) {
		for (VirtualFile f : files) {
			BzrVcs vcs = getVCS(f);
			if (vcs != null) { return vcs; }
		}
		return null;
	}

	@Nullable
	static private BzrVcs getVCS(VirtualFile file) {
	  for (Project project : ProjectManager.getInstance().getOpenProjects()) {
		 AbstractVcs vcs = ProjectLevelVcsManager.getInstance(project).getVcsFor(file);
		 if (vcs instanceof BzrVcs) {
			Log.info("Found BzrVcs for file %s: %s", file, vcs); 
			return (BzrVcs) vcs;
		 }
	  }
	  return null;
	}

}
