package bzr4idea;

import static bzr4idea.Log.info;
import com.intellij.openapi.Disposable;
import com.intellij.openapi.components.ServiceManager;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.vcs.AbstractVcs;
import com.intellij.openapi.vcs.CommittedChangesProvider;
import com.intellij.openapi.vcs.ProjectLevelVcsManager;
import com.intellij.openapi.vcs.VcsConfiguration;
import com.intellij.openapi.vcs.VcsShowConfirmationOption;
import com.intellij.openapi.vcs.annotate.AnnotationProvider;
import com.intellij.openapi.vcs.checkin.CheckinEnvironment;
import com.intellij.openapi.vcs.diff.DiffProvider;
import com.intellij.openapi.vcs.history.VcsHistoryProvider;
import com.intellij.openapi.vcs.rollback.RollbackEnvironment;
import com.intellij.openapi.vfs.LocalFileSystem;
import com.intellij.openapi.vfs.VirtualFile;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.Nullable;
import org.vcs.bazaar.client.BazaarClientPreferences;
import org.vcs.bazaar.client.BazaarPreference;
import org.vcs.bazaar.client.BazaarRevision;
import org.vcs.bazaar.client.IBazaarClient;
import org.vcs.bazaar.client.IBazaarNotifyListener;
import org.vcs.bazaar.client.BazaarClientFactory;
import org.vcs.bazaar.client.core.BazaarClientException;
import org.vcs.bazaar.client.commandline.CommandLineNotificationHandler;
import org.vcs.bazaar.client.xmlrpc.XmlRpcClientFactory;

import java.io.File;
import java.util.List;

/**
 * @author Patrik Beno
 */
public class BzrVcs extends AbstractVcs implements Disposable {

	public static final String VCS_NAME = "Bazaar";

	private Project project;
	private BzrLocalFileOperationsHandler localFileOperationsHandler;
	private AnnotationProvider annotationProvider;
	private VcsHistoryProvider historyProvider;
	private BzrChangeProvider changeProvider;
	private RollbackEnvironment rollbackEnvironment;
	private CheckinEnvironment checkinEnvironment;
	private CommittedChangesProvider committedChangesProvider;
	private VcsShowConfirmationOption addConfirmation;

	private IBazaarClient bzrclient;
	private BzrVcsConfigurable configurable;


	public BzrVcs(Project project) {
		super(project);

		this.project = project;
		ProjectLevelVcsManager vcsManager = ProjectLevelVcsManager.getInstance(project);
		addConfirmation = vcsManager
				.getStandardConfirmation(VcsConfiguration.StandardConfirmation.ADD, this);

		localFileOperationsHandler = new BzrLocalFileOperationsHandler();
	}

	@NonNls
	public String getName() {
		return VCS_NAME;
	}

	@NonNls
	public String getDisplayName() {
		return getName();
	}

	public BzrVcsConfigurable getConfigurable() {
		if (configurable == null) {
			configurable = new BzrVcsConfigurable(getProject());
		}
		return configurable;
	}

	public void activate() {
		info("BzrVcs.activate()");
		super.activate();

		LocalFileSystem.getInstance().registerAuxiliaryFileOperationsHandler(localFileOperationsHandler);

		CommandLineNotificationHandler notificationHandler = new CommandLineNotificationHandler();
		notificationHandler.enableLog();
		notificationHandler.add(new IBazaarNotifyListener() {
			public void logMessage(String message) {
				Log.info(message);
			}

			public void logError(String message) {
				Log.info(message);
			}

			public void logRevision(long revision, String path) {
				Log.info("revision: %s, path: %s", revision, path);
			}

			public void logCommandOutput(String message) {
				Log.bzrlog(message);
			}

			public void logCompleted(String message) {
				Log.info("completed: %s", message);
			}

			public void setCommand(int command) {
				Log.info("setCommand(%s)", command);
			}

			public void logCommandLine(String commandLine) {
				Log.info("command line: %s", commandLine);
			}

			public void onNotify(File file) {
				Log.info("onNotify(%s)", file); 
			}
		});

		BzrVcsSettings settings = ServiceManager.getService(BzrVcsSettings.class);

		BazaarClientPreferences.getInstance().set(
				BazaarPreference.EXECUTABLE,
				settings.getBzrExecutable());

		try {
			XmlRpcClientFactory.setup(true);
			BazaarClientFactory.setPreferredClientType(XmlRpcClientFactory.CLIENT_TYPE);
			bzrclient = BazaarClientFactory.createClient(XmlRpcClientFactory.CLIENT_TYPE);
		} catch (BazaarClientException e) {
			throw Assert.notYetHandled(e);
		}

	}

	public void deactivate() {
		info("BzrVcs.deactivate()");
		LocalFileSystem.getInstance().unregisterAuxiliaryFileOperationsHandler(localFileOperationsHandler);
		super.deactivate();
	}

	public boolean isVersionedDirectory(VirtualFile dir) {
		final VirtualFile versionFile = dir.findChild( ".bzr" );
		return versionFile != null && versionFile.isDirectory();
	}

	@Nullable
	// todo: implement this method to make possible operations: compare file with last revision
	public DiffProvider getDiffProvider() {
		return null;
	}

	@Nullable
	public AnnotationProvider getAnnotationProvider() {
		if (annotationProvider == null) {
			annotationProvider = new BzrAnnotationProvider(this);
		}

		return annotationProvider;
	}

	@Nullable
	public VcsHistoryProvider getVcsHistoryProvider() {
		if (historyProvider == null) {
			historyProvider = new BzrHistoryProvider(this);
		}
		return historyProvider;
	}

	@Nullable
	public VcsHistoryProvider getVcsBlockHistoryProvider() {
		return getVcsHistoryProvider();
	}

	@Nullable
	public BzrChangeProvider getChangeProvider() {
		if (changeProvider == null) {
			changeProvider = new BzrChangeProvider(this);
		}

		return changeProvider;
	}

	@Nullable
	public RollbackEnvironment getRollbackEnvironment() {
		if (rollbackEnvironment == null) {
			rollbackEnvironment = new BzrRollbackEnvironment(this);
		}

		return rollbackEnvironment;
	}

	@Nullable
	public CheckinEnvironment getCheckinEnvironment() {
		if (checkinEnvironment == null) {
			checkinEnvironment = new BzrCheckinEnvironment(this);
		}
		return checkinEnvironment;
	}

	@Nullable
	public BzrRevisionNumber parseRevisionNumber(String revisionNumberString) {
		int idx = revisionNumberString.indexOf(':');
		String prefix = (idx > 0) ? revisionNumberString.substring(0, idx + 1) : "revno:";
		String revision = (idx > 0) ? revisionNumberString.substring(idx + 1) : revisionNumberString;
		return new BzrRevisionNumber(BazaarRevision.getRevision(prefix, revision));
	}

	public CommittedChangesProvider getCommittedChangesProvider() {
		if (committedChangesProvider == null) {
			committedChangesProvider = new BzrCommittedChangesProvider(this);
		}
		return committedChangesProvider;
	}

	public Project getProject() {
		return project;
	}

	public IBazaarClient getBzrClient() {
		return bzrclient;
	}

	public VcsShowConfirmationOption getAddConfirmation() {
		return addConfirmation;
	}

	public void dispose() {
	}

	public void onSuccessfulCommit(List<File> files) {
		getChangeProvider().onSuccessfulCommit(files);
	}
}
